from django.contrib import admin
from django.urls import include, path

from huscy.projects.urls import router


urlpatterns = [
    path('admin/', admin.site.urls),
    path('api-auth/', include('rest_framework.urls')),
    path('api/', include(router.urls)),
]



from rest_framework import serializers

from huscy.projects.models import ResearchUnit


class ResearchUnitSerializer(serializers.ModelSerializer):
    class Meta:
        model = ResearchUnit
        fields = (
            'code',
            'id',
            'name',
            'principal_investigator',
        )



from rest_framework.routers import DefaultRouter
from rest_framework_nested.routers import NestedDefaultRouter

from huscy.projects import views


router = DefaultRouter()
router.register('projects', views.ProjectViewSet, basename='project')
router.register('researchunits', views.ResearchUnitViewSet)

project_router = NestedDefaultRouter(router, 'projects', lookup='project')
project_router.register('memberships', views.MembershipViewSet, basename='membership')

urlpatterns = router.urls
urlpatterns += project_router.urls



from django.contrib import admin

from huscy.projects import models


class ResearchUnitAdmin(admin.ModelAdmin):
    list_display = 'name', 'code', 'principal_investigator'


admin.site.register(models.Project)
admin.site.register(models.ResearchUnit, ResearchUnitAdmin)



# helper function to get ip
def get_client_ip(request):
    x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
    if x_forwarded_for:
        ip = x_forwarded_for.split(',')[0]
    else:
        ip = request.META.get('REMOTE_ADDR')
    return ip



from model_bakery import baker

from huscy.projects.models import Membership, Project
from huscy.projects.services import create_project


def test_creator_is_principal_investigator(user, research_unit):
    project = create_project('title', research_unit, user, user)

    assert Project.objects.exists()
    assert Membership.objects.count() == 1
    assert Membership.objects.filter(user=user, is_coordinator=True).exists()
    assert project.description == ''
    assert user.has_perm('change_project', project)


def test_creator_is_not_principal_investigator(django_user_model, user, research_unit):
    principal_investigator = baker.make(django_user_model)
    project = create_project('title', research_unit, principal_investigator, creator=user)

    assert Project.objects.exists()
    assert Membership.objects.count() == 2
    assert Membership.objects.filter(user=user, is_coordinator=True).exists()
    assert Membership.objects.filter(user=principal_investigator, is_coordinator=True).exists()
    assert project.description == ''
    assert user.has_perm('change_project', project)
    assert principal_investigator.has_perm('change_project', project)


def test_with_optional_description(user, research_unit):
    project = create_project('title', research_unit, user, user, description='description')

    assert Project.objects.exists()
    assert not project.description == ''


def test_with_optional_local_id(user, research_unit):
    project = create_project('title', research_unit, user, user, local_id=166)
    assert Project.objects.exists()
    assert project.local_id == 166



from operator import attrgetter

import pytest
from model_bakery import baker

from huscy.projects.services import get_projects

pytestmark = pytest.mark.django_db


@pytest.fixture
def projects():
    return baker.make('projects.Project', _quantity=10)


def test_get_projects(projects):
    result = get_projects()

    assert list(sorted(projects, key=attrgetter('id'), reverse=True)) == list(result)



from datetime import timedelta

import pytest
from model_bakery import baker

from rest_framework.test import APIClient

from huscy.projects.models import Project


ONE_HOUR = timedelta(hours=1)


@pytest.fixture
def user(django_user_model):
    return django_user_model.objects.create_user(username='user', password='password',
                                                 first_name='Phil', last_name='Stift')


@pytest.fixture
def admin_client(admin_user):
    client = APIClient()
    client.login(username=admin_user.username, password='password')
    return client


@pytest.fixture
def client(user):
    client = APIClient()
    client.login(username=user.username, password='password')
    return client


@pytest.fixture
def anonymous_client():
    return APIClient()


@pytest.fixture
def research_unit(user):
    return baker.make(
        'projects.ResearchUnit',
        name='research_unit_1',
        code='RU1',
        principal_investigator=user,
    )


@pytest.fixture
def project():
    return baker.make(Project, title='project_title')


@pytest.fixture
def experiment(project):
    return baker.make('projects.Experiment', project=project, description='description')


@pytest.fixture
def session(experiment):
    return baker.make(
        'projects.Session',
        experiment=experiment,
        duration=ONE_HOUR,
    )


@pytest.fixture
def data_acquisition_method(session):
    return baker.make('projects.DataAcquisitionMethod', session=session)


@pytest.fixture
def membership(project):
    return baker.make('projects.Membership', project=project)



# Generated by Django 3.2.8 on 2021-10-25 05:41

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import huscy.projects.models.reserach_unit


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='ResearchUnit',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=128, unique=True)),
                ('code', models.CharField(blank=True, default=huscy.projects.models.reserach_unit.get_default_value_for_code, max_length=32, unique=True)),
                ('principal_investigator', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='Project',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('local_id', models.PositiveIntegerField()),
                ('title', models.CharField(max_length=126)),
                ('description', models.TextField(blank=True, default='')),
                ('principal_investigator', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL)),
                ('research_unit', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, to='projects.researchunit')),
            ],
            options={
                'ordering': ('-id',),
                'unique_together': {('local_id', 'research_unit')},
            },
        ),
        migrations.CreateModel(
            name='Membership',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('is_coordinator', models.BooleanField(default=False)),
                ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='projects.project')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'unique_together': {('user', 'project')},
            },
        ),
    ]



from model_bakery import baker

from huscy.projects import services


def test_no_project_participation(admin_user):
    users_project = services.get_participating_projects(admin_user).values_list('id', flat=True)
    assert list(users_project) == []


def test_pi_participated_projects(user, admin_user):
    pi_project = baker.make('projects.Project', principal_investigator=user)
    user_projects = services.get_participating_projects(user).values_list('id', flat=True)
    assert list(user_projects) == [pi_project.id]


def test_member_projects(user, project):
    user_projects = services.get_participating_projects(user).values_list('id', flat=True)
    assert list(user_projects) == []

    services.add_member(project, user)

    user_projects = services.get_participating_projects(user).values_list('id', flat=True)
    assert list(user_projects) == [project.id]


def test_pi_and_member_projects(user):
    pi_project = baker.make('projects.Project', principal_investigator=user)
    member_project = baker.make('projects.Project')

    services.add_member(member_project, user)

    user_projects = services.get_participating_projects(user).values_list('id', flat=True)
    assert sorted(user_projects) == sorted([pi_project.id, member_project.id])


def test_pi_and_member_are_the_same(user):
    pi_project = baker.make('projects.Project', principal_investigator=user)
    services.add_member(pi_project, user)

    user_projects = services.get_participating_projects(user).values_list('id', flat=True)
    assert list(user_projects) == [pi_project.id]



from .membership import MembershipSerializer
from .project import ProjectSerializer
from .research_unit import ResearchUnitSerializer


__all__ = (
    'MembershipSerializer',
    'ProjectSerializer',
    'ResearchUnitSerializer',
)



import pytest

from rest_framework.reverse import reverse
from rest_framework.status import HTTP_204_NO_CONTENT, HTTP_403_FORBIDDEN

from django.contrib.auth.models import Permission

from huscy.projects import services

pytestmark = pytest.mark.django_db


def test_admin_user_can_delete_projects(admin_client, project):
    response = delete_project(admin_client, project)

    assert response.status_code == HTTP_204_NO_CONTENT


def test_user_with_permission_can_delete_projects(client, user, project):
    delete_permission = Permission.objects.get(codename='delete_project')
    user.user_permissions.add(delete_permission)

    response = delete_project(client, project)

    assert response.status_code == HTTP_204_NO_CONTENT


def test_user_without_permission_cannot_delete_projects(client, project):
    response = delete_project(client, project)

    assert response.status_code == HTTP_403_FORBIDDEN


def test_anonymous_user_cannot_delete_projects(anonymous_client, project):
    response = delete_project(anonymous_client, project)

    assert response.status_code == HTTP_403_FORBIDDEN


@pytest.mark.skip
def test_service_function_was_called(mocker, client, user, project):
    spy = mocker.spy(services, 'delete_project')

    delete_permission = Permission.objects.get(codename='delete_project')
    user.user_permissions.add(delete_permission)

    delete_project(client, project)

    spy.assert_called_once()


def delete_project(client, project):
    return client.delete(reverse('project-detail', kwargs=dict(pk=project.pk)))



import pytest
from model_bakery import baker

from rest_framework.reverse import reverse
from rest_framework.status import HTTP_201_CREATED, HTTP_400_BAD_REQUEST, HTTP_403_FORBIDDEN

pytestmark = pytest.mark.django_db


def test_admin_user_can_create_projects(admin_client, admin_user, research_unit):
    response = create_project(admin_client, research_unit, admin_user)

    assert response.status_code == HTTP_201_CREATED


def test_user_can_create_projects(client, user, research_unit):
    response = create_project(client, research_unit, user)

    assert response.status_code == HTTP_201_CREATED


def test_anonymous_user_cannot_create_projects(anonymous_client, user, research_unit):
    response = create_project(anonymous_client, research_unit, user)

    assert response.status_code == HTTP_403_FORBIDDEN


def test_create_project_with_existing_local_id(client, user, research_unit):
    baker.make('projects.Project', research_unit=research_unit, local_id=166)
    response = create_project(client, research_unit, user, local_id=166)

    assert response.status_code == HTTP_400_BAD_REQUEST
    error_msg = response.json()['non_field_errors']
    assert error_msg == ['The fields local_id, research_unit must make a unique set.']


def create_project(client, research_unit, principal_investigator=None, local_id=None):
    data = dict(
        description='project_description',
        principal_investigator=principal_investigator.pk,
        research_unit=research_unit.id,
        title='project_title',
    )
    if local_id:
        data['local_id'] = local_id

    return client.post(reverse('project-list'), data=data)



import pytest

from huscy.projects.models import Membership
from huscy.projects.services import add_member, remove_member

pytestmark = pytest.mark.django_db


def test_remove_membership(project, user):
    membership = add_member(project, user)  # use service function here to set the permission
    assert_user_has_view_permission_for_project(project, user)
    assert_user_does_not_have_change_permission_for_project(project, user)

    remove_member(membership)

    assert_user_does_not_have_view_permission_for_project(project, user)
    assert_user_does_not_have_change_permission_for_project(project, user)
    assert_membership_does_not_exist()


def test_write_permission_is_removed_as_well(project, user):
    membership = add_member(project, user, is_coordinator=True)  # user has write permission now
    assert_user_has_change_permission_for_project(project, user)

    remove_member(membership)

    assert_user_does_not_have_view_permission_for_project(project, user)
    assert_user_does_not_have_change_permission_for_project(project, user)
    assert_membership_does_not_exist()


def assert_membership_does_not_exist():
    assert not Membership.objects.exists()


def assert_membership_exists(project, user):
    membership = Membership.objects.get()
    assert membership.project == project
    assert membership.user == user


def assert_user_has_view_permission_for_project(project, user):
    assert user.has_perm('view_project', project)


def assert_user_does_not_have_view_permission_for_project(project, user):
    assert not user.has_perm('view_project', project)


def assert_user_has_change_permission_for_project(project, user):
    assert user.has_perm('change_project', project)


def assert_user_does_not_have_change_permission_for_project(project, user):
    assert not user.has_perm('change_project', project)



from model_bakery import baker


def test_str_method(django_user_model):
    user = baker.prepare(django_user_model, first_name='first_name', last_name='last_name')
    research_unit = baker.prepare('projects.ResearchUnit', name='ResearchUnit1', code='RU1',
                                  principal_investigator=user)

    assert 'ResearchUnit1 (first_name last_name)' == str(research_unit)



from rest_framework.permissions import BasePermission, SAFE_METHODS
from huscy.projects.models import Membership


class AllowAnyToCreate(BasePermission):
    def has_permission(self, request, view):
        return request.method == 'POST'


class IsProjectCoordinator(BasePermission):
    def has_permission(self, request, view):
        return Membership.objects.filter(project=view.project, user=request.user,
                                         is_coordinator=True).exists()


class IsProjectMember(BasePermission):
    def has_permission(self, request, view):
        return Membership.objects.filter(project=view.project, user=request.user).exists()


class ReadOnly(BasePermission):
    def has_permission(self, request, view):
        return request.method in SAFE_METHODS



import pytest

from huscy.projects.serializer import ProjectSerializer

pytestmark = pytest.mark.django_db


def test_project_serializer(rf, user, project):
    request = rf.get('/any/path/')
    request.user = user

    expected = dict(
        description=project.description,
        id=project.id,
        local_id=project.local_id,
        local_id_name=project.local_id_name,
        participating=False,
        principal_investigator=project.principal_investigator.pk,
        principal_investigator_name=project.principal_investigator.get_full_name(),
        research_unit=project.research_unit.pk,
        research_unit_name=project.research_unit.name,
        title=project.title,
    )

    assert expected == ProjectSerializer(instance=project, context={'request': request}).data


def test_participating_flag_with_member(rf, membership):
    request = rf.get('/any/path/')
    request.user = membership.user

    serializer = ProjectSerializer(instance=membership.project, context={'request': request})

    assert serializer.data['participating'] is True


def test_participating_flag_with_principal_investigator(rf, project):
    request = rf.get('/any/path/')
    request.user = project.principal_investigator

    serializer = ProjectSerializer(instance=project, context={'request': request})

    assert serializer.data['participating'] is True


def test_participating_flag_with_any_other_user(rf, user, project):
    request = rf.get('/any/path/')
    request.user = user

    serializer = ProjectSerializer(instance=project, context={'request': request})

    assert serializer.data['participating'] is False



from rest_framework import serializers

from huscy.projects.models import Membership
from huscy.projects.services import add_member, update_membership


class MembershipSerializer(serializers.ModelSerializer):
    has_write_permission = serializers.BooleanField()
    username = serializers.SerializerMethodField()

    class Meta:
        model = Membership
        fields = (
            'id',
            'has_write_permission',
            'is_coordinator',
            'project',
            'user',
            'username',
        )
        read_only_fields = 'project',

    def get_username(self, membership):
        return membership.user.get_full_name()

    def create(self, validated_data):
        project = self.context['project']
        return add_member(project=project, **validated_data)

    def update(self, membership, validated_data):
        return update_membership(membership, validated_data.get('is_coordinator'),
                                 validated_data.get('has_write_permission'))



from model_bakery import baker

from huscy.projects.serializer import MembershipSerializer
from huscy.projects.services import add_member


def test_project_serializer(django_user_model, project):
    user = baker.make(django_user_model, first_name='first_name', last_name='last_name')
    membership = add_member(project, user, is_coordinator=True)

    expected = {
        'id': membership.id,
        'is_coordinator': True,
        'has_write_permission': True,
        'project': project.id,
        'user': membership.user.pk,
        'username': 'first_name last_name',
    }

    assert expected == MembershipSerializer(instance=membership).data



from itertools import cycle

from model_bakery import baker

from huscy.projects.services import get_memberships


def test_get_memberships(django_user_model):
    projects = baker.make('projects.Project', _quantity=3)
    users = baker.make(django_user_model, _quantity=3)
    baker.make('projects.Membership', project=cycle(projects), user=cycle(users), _quantity=3)

    memberships = get_memberships()

    assert len(memberships) == 3


def test_get_memberships_filtered_by_project(django_user_model):
    projects = baker.make('projects.Project', _quantity=3)
    users = baker.make(django_user_model, _quantity=3)
    baker.make('projects.Membership', project=cycle(projects), user=cycle(users), _quantity=3)

    project = projects[0]
    memberships = get_memberships(project)

    assert len(memberships) == 1
    assert project == memberships[0].project



import logging

from django.shortcuts import get_object_or_404
from rest_framework import mixins, viewsets
from rest_framework.permissions import DjangoModelPermissions, IsAuthenticated, IsAdminUser

from huscy.projects import helpers, serializer, services
from huscy.projects.models import Project
from huscy.projects.permissions import AllowAnyToCreate, IsProjectCoordinator, ReadOnly

logger = logging.getLogger('projects')


class MembershipViewSet(mixins.CreateModelMixin, mixins.DestroyModelMixin, mixins.ListModelMixin,
                        mixins.UpdateModelMixin, viewsets.GenericViewSet):
    serializer_class = serializer.MembershipSerializer
    permission_classes = (IsAuthenticated, IsAdminUser | IsProjectCoordinator | ReadOnly)

    def initial(self, request, *args, **kwargs):
        self.project = get_object_or_404(Project, pk=self.kwargs['project_pk'])
        super().initial(request, *args, **kwargs)

    def get_queryset(self):
        return services.get_memberships(self.project)

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context['project'] = self.project
        return context

    def perform_destroy(self, membership):
        services.remove_member(membership)


class ProjectViewSet(mixins.CreateModelMixin, mixins.DestroyModelMixin, mixins.ListModelMixin,
                     mixins.UpdateModelMixin, viewsets.GenericViewSet):
    serializer_class = serializer.ProjectSerializer
    permission_classes = (IsAuthenticated, DjangoModelPermissions | AllowAnyToCreate)

    def get_queryset(self):
        return services.get_projects()

    def destroy(self, request, *args, **kwargs):
        pk = kwargs.get('pk')
        logger.info('User %s from ip %s requested deletion of project with id:%d',
                    request.user.username, helpers.get_client_ip(request), pk)
        return super().destroy(request, args, kwargs)

    def perform_destroy(self, project):
        services.delete_project(project)


class ResearchUnitViewSet(viewsets.ModelViewSet):
    permission_classes = (DjangoModelPermissions, )
    queryset = services.get_research_units()
    serializer_class = serializer.ResearchUnitSerializer



import pytest
from model_bakery import baker

from huscy.projects import models

pytestmark = pytest.mark.django_db


def test_get_next_local_id_on_empty_projects(research_unit):
    assert not models.Project.objects.exists()
    assert models.Project.objects.get_next_local_id(research_unit) == 1


def test_get_next_local_id_with_one_project():
    project = baker.make('projects.Project', local_id=1)
    assert models.Project.objects.count() == 1
    assert models.Project.objects.get_next_local_id(project.research_unit) == 2


def test_get_next_local_id_with_gap_in_local_ids(research_unit):
    baker.make('projects.Project', local_id=1, research_unit=research_unit)
    baker.make('projects.Project', local_id=3, research_unit=research_unit)
    assert models.Project.objects.get_next_local_id(research_unit) == 2



import pytest
from model_bakery import baker

from rest_framework.reverse import reverse
from rest_framework.status import (
    HTTP_200_OK,
    HTTP_201_CREATED,
    HTTP_204_NO_CONTENT,
    HTTP_403_FORBIDDEN,
    HTTP_405_METHOD_NOT_ALLOWED,
)

pytestmark = pytest.mark.django_db


@pytest.fixture
def member(django_user_model):
    return baker.make(django_user_model)


def test_retrieve_membership_is_not_provided(client, project, membership):
    response = client.get(
        reverse('membership-detail', kwargs=dict(project_pk=project.pk, pk=membership.pk))
    )

    assert response.status_code == HTTP_405_METHOD_NOT_ALLOWED


def test_admin_user_can_create_memberships(admin_client, project, member):
    response = create_membership(admin_client, project, member)

    assert response.status_code == HTTP_201_CREATED


def test_admin_user_can_delete_memberships(admin_client, project, membership):
    response = delete_membership(admin_client, project, membership)

    assert response.status_code == HTTP_204_NO_CONTENT


def test_admin_user_can_list_memberships(admin_client, project):
    response = list_memberships(admin_client, project)

    assert response.status_code == HTTP_200_OK


def test_admin_user_can_update_memberships(admin_client, project, membership):
    response = update_membership(admin_client, project, membership)

    assert response.status_code == HTTP_200_OK


@pytest.mark.parametrize('is_coordinator,expected_status_code', [
    (True, HTTP_201_CREATED),
    (False, HTTP_403_FORBIDDEN),
])
def test_create_membership_when_user_has_membership(client, user, project, member,
                                                    is_coordinator, expected_status_code):
    baker.make('projects.Membership', project=project, user=user, is_coordinator=is_coordinator)

    response = create_membership(client, project, member)

    assert response.status_code == expected_status_code


def test_user_without_membership_cannot_create_memberships(client, user, project, member):
    response = create_membership(client, project, member)

    assert response.status_code == HTTP_403_FORBIDDEN


@pytest.mark.parametrize('is_coordinator,expected_status_code', [
    (True, HTTP_204_NO_CONTENT),
    (False, HTTP_403_FORBIDDEN),
])
def test_delete_membership_with_project_membership(client, user, project, membership,
                                                   is_coordinator, expected_status_code):
    baker.make('projects.Membership', project=project, user=user, is_coordinator=is_coordinator)

    response = delete_membership(client, project, membership)

    assert response.status_code == expected_status_code


def test_user_without_membership_cannot_delete_memberships(client, user, project, membership):
    response = delete_membership(client, project, membership)

    assert response.status_code == HTTP_403_FORBIDDEN


def test_user_without_membership_can_list_memberships(client, project):
    response = list_memberships(client, project)

    assert response.status_code == HTTP_200_OK


@pytest.mark.parametrize('is_coordinator,expected_status_code', [
    (True, HTTP_200_OK),
    (False, HTTP_403_FORBIDDEN),
])
def test_update_membership_with_project_membership(client, user, project, membership,
                                                   is_coordinator, expected_status_code):
    baker.make('projects.Membership', project=project, user=user, is_coordinator=is_coordinator)

    response = update_membership(client, project, membership)

    assert response.status_code == expected_status_code


def test_user_without_membership_can_update_memberships(client, user, project, membership):
    response = update_membership(client, project, membership)

    assert response.status_code == HTTP_403_FORBIDDEN


def test_anonymous_user_cannot_create_memberships(anonymous_client, project, member):
    response = create_membership(anonymous_client, project, member)

    assert response.status_code == HTTP_403_FORBIDDEN


def test_anonymous_user_cannot_delete_memberships(anonymous_client, project, membership):
    response = delete_membership(anonymous_client, project, membership)

    assert response.status_code == HTTP_403_FORBIDDEN


def test_anonymous_user_cannot_list_memberships(anonymous_client, project):
    response = list_memberships(anonymous_client, project)

    assert response.status_code == HTTP_403_FORBIDDEN


def test_anonymous_user_cannot_update_memberships(anonymous_client, project, membership):
    response = update_membership(anonymous_client, project, membership)

    assert response.status_code == HTTP_403_FORBIDDEN


def create_membership(client, project, member, is_coordinator=False):
    return client.post(
        reverse('membership-list', kwargs=dict(project_pk=project.pk)),
        data=dict(user=member.id, is_coordinator=is_coordinator)
    )


def delete_membership(client, project, membership):
    return client.delete(
        reverse('membership-detail', kwargs=dict(project_pk=project.pk, pk=membership.id))
    )


def list_memberships(client, project):
    return client.get(reverse('membership-list', kwargs=dict(project_pk=project.pk)))


def update_membership(client, project, membership):
    return client.put(
        reverse('membership-detail', kwargs=dict(project_pk=project.pk, pk=membership.id)),
        data=dict(
            user=membership.user.id,
            is_coordinator=False,
            has_write_permission=True,
        ),
    )



from .memberships import (add_member, get_memberships, get_participating_projects, remove_member,
                          update_membership)
from .projects import create_project, delete_project, get_projects
from .research_units import get_research_units


__all__ = (
    'add_member',
    'create_project',
    'delete_project',
    'get_memberships',
    'get_participating_projects',
    'get_projects',
    'get_research_units',
    'remove_member',
    'update_membership',
)



from model_bakery import baker


def test_str_method(django_user_model):
    project = baker.prepare('projects.Project', title='project1')
    user = baker.prepare(django_user_model, first_name='first_name', last_name='last_name')
    membership = baker.prepare('projects.Membership', project=project, user=user)

    expected_result = 'first_name last_name is member in project project1'
    assert expected_result == str(membership)



import pytest

from rest_framework.reverse import reverse
from rest_framework.status import HTTP_200_OK, HTTP_403_FORBIDDEN

pytestmark = pytest.mark.django_db


def test_admin_user_can_view_projects(admin_client):
    response = list_projects(admin_client)

    assert response.status_code == HTTP_200_OK


def test_user_without_permission_can_view_projects(client):
    response = list_projects(client)

    assert response.status_code == HTTP_200_OK


def test_anonymous_user_cannot_view_projects(anonymous_client):
    response = list_projects(anonymous_client)

    assert response.status_code == HTTP_403_FORBIDDEN


def list_projects(client):
    return client.get(reverse('project-list'))


def retrieve_project(client, project):
    return client.get(reverse('project-detail', kwargs=dict(pk=project.pk)))



from itertools import count, filterfalse

from django.conf import settings
from django.db import models

from .reserach_unit import ResearchUnit


class ProjectManager(models.Manager):
    def get_next_local_id(self, research_unit):
        taken_values = self.filter(research_unit=research_unit).values_list('local_id', flat=True)
        return next(filterfalse(lambda x: x in set(taken_values), count(1)))


class Project(models.Model):
    research_unit = models.ForeignKey(ResearchUnit, on_delete=models.DO_NOTHING)
    local_id = models.PositiveIntegerField()

    title = models.CharField(max_length=126)

    description = models.TextField(blank=True, default='')

    principal_investigator = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.PROTECT)

    objects = ProjectManager()

    @property
    def local_id_name(self):
        return f'{self.research_unit.code}-{self.local_id}'

    class Meta:
        ordering = '-id',
        unique_together = ('local_id', 'research_unit')



from os import path

from setuptools import find_namespace_packages, setup

from huscy.projects import __version__


with open(path.join(path.abspath(path.dirname(__file__)), 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

extras_require = {
    'development': [
        'psycopg2-binary',
    ],
    'testing': [
        'tox',
        'watchdog==0.9',
    ],
}

install_requires = [
    'Django>=2.2',
    'djangorestframework>=3.10',
    'django-guardian>=1.4',
    'drf-nested-routers>=0.90',
]


setup(
    name='huscy.projects',
    version=__version__,
    license='AGPLv3+',

    author='Alexander Tyapkov, Mathias Goldau, Stefan Bunde',
    author_email='tyapkov@gmail.com, goldau@cbs.mpg.de, stefanbunde+git@posteo.de',

    packages=find_namespace_packages(),

    install_requires=install_requires,
    extras_require=extras_require,

    long_description=long_description,
    long_description_content_type='text/markdown',

    url='https://huscy.org',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Framework :: Django',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.0',
        'Framework :: Django :: 3.1',
    ],
)



from django.conf import settings
from django.db import models

from .project import Project


class Membership(models.Model):
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    project = models.ForeignKey(Project, on_delete=models.CASCADE)
    is_coordinator = models.BooleanField(default=False)

    def has_write_permission(self):
        return self.user.has_perm('change_project', self.project)

    def __str__(self):
        return f'{self.user.get_full_name()} is member in project {self.project.title}'

    class Meta:
        unique_together = 'user', 'project'
